from io import BytesIO
from pathlib import Path
from typing import Optional, Union

Bytes = Union[bytes, BytesIO]
FilePath = Union[str, Path]


class File:
    try:
        from aiofiles import open
    except ImportError as e:
        raise SystemExit(
            "You must install `aiofiles` package to be able to use File util."
        ) from e

    def __init__(
        self, name: Optional[str] = None, file_source: Optional["Bytes"] = None
    ) -> None:
        self.name, self.file_source = name, file_source

    @classmethod
    def from_bytes(cls, file_source: Bytes, name: Optional[str] = None) -> "File":
        file_source = (
            file_source.getvalue() if isinstance(file_source, BytesIO) else file_source
        )
        return cls(name=name, file_source=file_source)

    @classmethod
    async def from_path(
        cls, file_source: FilePath, name: Optional[str] = None
    ) -> "File":
        path = Path(file_source) if isinstance(file_source, str) else path

        async with cls.open(path, "rb") as f:
            return cls(name=name or path.name, file_source=await f.read())

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r}>"
