from re import compile
from inspect import iscoroutinefunction
from typing import Iterable, List, Optional, Type, Union

from telefone.bot.states import BaseStateGroup, get_state_repr
from telefone.bot.updates import CallbackQueryUpdate, MessageUpdate

from telefone.bot.rules.abc import ABCRule


class CallbackData(ABCRule[MessageUpdate]):
    """
    A rule that checks if callback query is equal to or is in the list of given data
    """

    def __init__(self, callback_data: Union[str, List[str]]):
        if isinstance(callback_data, str):
            callback_data = [callback_data]

        self.callback_data = callback_data

    async def check(self, callback_query: CallbackQueryUpdate) -> bool:
        return callback_query.data in self.callback_data


class Command(ABCRule[MessageUpdate]):
    def __init__(
        self, text: str, prefixes: Optional[Union[str, Iterable[str]]] = "/"
    ) -> None:
        if isinstance(prefixes, str):
            prefixes = [prefixes]

        self.text = text
        self.prefixes = prefixes

    async def check(self, message: MessageUpdate) -> bool:
        for prefix in self.prefixes:
            if not message.text.startswith(prefix):
                continue
            elif message.text.startswith(prefix + self.text + " "):
                return {"args": message.text[len(prefix + self.text) + 1 :].split()}
            elif message.text == prefix + self.text:
                return True
        return False


class Function(ABCRule[MessageUpdate]):
    def __init__(self, func) -> None:
        self.func = func

    async def check(self, message: MessageUpdate) -> bool:
        if iscoroutinefunction(self.func):
            return await self.func(message)


class IsPrivate(ABCRule[MessageUpdate]):
    """
    A rule that checks if the message is private.
    """

    async def check(self, message: MessageUpdate) -> bool:
        return message.chat.type == "private"


class IsGroup(ABCRule[MessageUpdate]):
    """
    A rule that checks if the message is from the chat.
    """

    async def check(self, message: MessageUpdate) -> bool:
        return message.chat.type in ["group", "supergroup"]


class Length(ABCRule[MessageUpdate]):
    """
    A rule that checks if the message is longer than or equal to the given length.
    """

    def __init__(self, minimum_length: int) -> None:
        self.minimum_length = minimum_length

    async def check(self, message: MessageUpdate) -> bool:
        return self.minimum_length >= len(message.text)


class Levenshtein(ABCRule[MessageUpdate]):
    def __init__(
        self, texts: Union[str, List[str]], max_distance: Optional[int] = 1
    ) -> None:
        if isinstance(texts, str):
            texts = [texts]

        self.texts = texts
        self.max_distance = max_distance

    @staticmethod
    def distance(a: str, b: str) -> int:
        n, m = len(a), len(b)
        if n > m:
            a, b = b, a
            n, m = m, n

        current_row = range(n + 1)
        for i in range(1, m + 1):
            previous_row, current_row = current_row, [i] + [0] * n  # type: ignore
            for j in range(1, n + 1):
                add, delete, change = (
                    previous_row[j] + 1,
                    current_row[j - 1] + 1,
                    previous_row[j - 1],
                )
                if a[j - 1] != b[i - 1]:
                    change += 1
                current_row[j] = min(add, delete, change)  # type: ignore

        return current_row[n]

    async def check(self, message: MessageUpdate) -> bool:
        message_text = message.text or message.caption

        if not message_text:
            return False
            
        for text in self.texts:
            if self.distance(text, message_text) <= self.max_distance:
                return True
        return False


class Regex(ABCRule[MessageUpdate]):
    """
    A rule that checks if the message matches the given regex.
    """

    def __init__(self, regex: str) -> None:
        self.regex = compile(regex)

    async def check(self, message: MessageUpdate) -> bool:
        return self.regex.match(message.text) is not None


class State(ABCRule[MessageUpdate]):
    def __init__(
        self,
        state: Optional[Union[BaseStateGroup, List[BaseStateGroup]]] = None,
    ):
        if not isinstance(state, list):
            state = [] if state is None else [state]
        self.state = [get_state_repr(s) for s in state]

    async def check(self, event: MessageUpdate) -> bool:
        if event.state_peer is None:
            return not self.state
        return event.state_peer.state in self.state


class StateGroup(ABCRule[MessageUpdate]):
    def __init__(
        self,
        state_group: Union[Type[BaseStateGroup], List[Type[BaseStateGroup]]],
    ):
        if not isinstance(state_group, list):
            state_group = [] if state_group is None else [state_group]
        self.state_group = state_group

    async def check(self, message: MessageUpdate) -> bool:
        if message.state_peer is None:
            return not self.state_group
        return type(message.state_peer.state) in self.state_group


class Text(ABCRule[MessageUpdate]):
    """
    A rule that checks if the message text is equal
    to or is in the list of given texts.
    """

    def __init__(
        self, texts: Union[str, List[str]], ignore_case: Optional[bool] = False
    ) -> None:
        if not isinstance(texts, list):
            texts = [texts]

        self.texts = texts
        self.ignore_case = ignore_case

    async def check(self, message: MessageUpdate) -> bool:
        if self.ignore_case:
            return message.text.lower() in [text.lower() for text in self.texts]
        return message.text in self.texts


class VBML(ABCRule[MessageUpdate]):
    try:
        from vbml import Patcher, Pattern
    except ImportError as e:
        raise SystemExit(
            "You must install `vbml` package to be able to use VBML rule."
        ) from e

    def __init__(
        self,
        patterns: Union[str, Pattern, Iterable[Union[str, Pattern]]],
        patcher: Optional[Patcher] = Patcher(),
    ) -> None:
        if not isinstance(patterns, list):
            patterns = [patterns]

        self.patterns = [
            self.Pattern(pattern) if isinstance(pattern, str) else pattern
            for pattern in patterns
        ]
        self.patcher = patcher

    async def check(self, message: MessageUpdate) -> bool:
        for pattern in self.patterns:
            result = self.patcher.check(pattern, message.text)
            if result not in (None, False):
                return result
        return False
